import { NextApiRequest, NextApiResponse } from "next"
import logging, { Logging } from "../../lib/backend/logging"
import getPlayerByNameDB from "../../lib/backend/components/getPlayerByNameDB"
import checkPasswordIsValid from "../../lib/backend/components/checkPasswordIsValid"
import createTokenDB, {
  IdToken,
  RawToken,
} from "../../lib/backend/components/createTokenDB"
import sendResponse from "../../lib/backend/components/sendResponse"
import { setCookie } from "cookies-next"
import { Player, Token } from "@prisma/client"
import prisma from "../../lib/prisma"
import jwt from "jsonwebtoken"
import errors from "../../lib/backend/errors"
import sendError, { API } from "../../lib/backend/components/sendError"
import getUserFromBody from "../../lib/backend/components/getUserFromBody"

interface Data {
  loggedIn: boolean
}

export default async function login(
  req: NextApiRequest,
  res: NextApiResponse<Data>
) {
  const context = { req, res }

  return preCheck(context, () =>
    getUserFromBody(context, (username, password) =>
      getPlayerByNameDB(context, username, (player) => {
        checkPasswordIsValid(context, player, password, () => {
          createTokenDB(player, "REFRESH", (newToken, newTokenDB) => {
            sendResponse(
              context,
              loginResponse(context, player, newToken, newTokenDB)
            )
          })
        })
      })
    )
  ).catch((err) => sendError(context, err))
}

async function preCheck(context: API, next: () => void) {
  const { req } = context
  const oldRefreshToken = req.cookies.token

  try {
    if (!oldRefreshToken) return sendError(context, errors.noCookie)
    // Check for old cookie, if unused invalidate it
    const tokenData = jwt.verify(
      oldRefreshToken,
      process.env.ACCESS_TOKEN_SECRET as string
    )

    if (!tokenData || typeof tokenData === "string") return

    const oldDBToken = await prisma.token.findUniqueOrThrow({
      where: {
        id: tokenData.id,
      },
    })

    if (oldDBToken.used)
      logging("Old login token was used: " + oldDBToken.id, ["debug"], req)
    else {
      await prisma.token.update({
        where: {
          id: tokenData.id,
        },
        data: {
          used: true,
        },
      })
      await logging("Old login token has been invalidated.", ["debug"], req)
    }
  } catch (err) {
    // err is expected if no correct cookie, just continue, otherwise it has been invalidated above
  }

  return next()
}

function loginResponse(
  context: API,
  player: Player,
  newToken: RawToken,
  newTokenDB: Token
) {
  const { req, res } = context
  // const { player, req, res } = payload
  const { value } = newToken

  // Set login cookie
  setCookie("token", value, {
    req,
    res,
    maxAge: 172800,
    httpOnly: true,
    sameSite: true,
    secure: true,
    path: "/api",
  })

  // Successfull response
  return {
    message:
      "User " +
      player.id +
      " logged in and generated Refresh-Token: " +
      newTokenDB.id,
    body: { loggedIn: true },
    type: ["debug", "info.cyan"] as Logging[],
  }
}
